/*
m_freqinverter - Shift frequencies of an audio signal around a centre freqeuncyt

DESCRIPTION
m_freqinverter shifts frequencies of an audio signal around a centre
frequency. Frequencies will be scaled to have the same proportions in the lower
and upper frequency bands. They will be shifted so, that the lowest frequency
of the lower band will become the lowest frequency of the upper band.
This opcode was designed with percussive sounds in mind, but can be used on any
type of signal to achieve quite unexpected results.

SYNTAX
aout m_freqinverter ain, kmin, kmax, kcenter, kcut, kbypass, ifftsize

INITIALIZATION
ifftsize - Size of the FFT. This must be a power-of-two (e.g. 512, 1024,
	2048, ...). A good starting value is 1024

PERFORMANCE
aout - The audio output signal of the opcode.
ain - The audio input signal to be processed
kmin - Lowest frequency to be considered
kmax - Highest frequency to be considered. Use sr/2 to have maximum upper limit
kcenter - Centre frequency to shift all other frequencies around
kcut - If not 0, cut out everything above kmax and below kmin
kbypass - Bypass the processing, if this is not zero

CREDITS
Author: Jeanette C., inspired by an idea by Set Hallstrm posted on the
Linux Audio Users' mailinglist.
*/

opcode m_freqinverter, a, akkkkki
ain, kmin, kmax, kcenter, kcut, kbypass, ifftsize xin

	ioverlap = ifftsize/64
	iwinsize = ifftsize
	iwintype = 1
	ifftres = sr / ifftsize
	ifftbins = (ifftsize / 2) + 1
	ifnf ftgen 0, 0, ifftsize, 7, 0 ; table allocated for input frequencies
	ifng ftgen 0, 0, ifftsize, 7, 0 ; Table allocated for output frequencies
	ifna ftgen 0, 0, ifftsize, 7, 0 ; Table allocted for input amplitudes
	ifnb ftgen 0, 0, ifftsize, 7, 0 ; Table allocated for output amplitudes

	; Delay the original signal to get smooth bypass crossing
	abuf delayr 0.1
	adel deltapn ifftsize
	delayw ain

	if kbypass == 0 then
		kgoto Process
	else
		aout = adel
		kgoto Output
	endif

Process:
	; Set the basic parameters of the opcode
	klowbw = kcenter - kmin
	khighbw = kmax - kcenter
	kfactor = khighbw / klowbw
	; FFT analyse the input and write it into a table
	fsig pvsanal ain, ifftsize, ioverlap, iwinsize, iwintype
	fsrc pvslock fsig, 1
	kflag pvsftw fsrc, ifna, ifnf

	if kflag == 1 then
		; We have analysis data, that we can process
		kndx = 0
		loop:
			kfreq tab kndx, ifnf
			kamp tab kndx, ifna
			; Calculate the new frequency from the current one
			if kfreq <= kcenter then
				if kfreq >= kmin then
					knewfreq = kcenter + kfactor * (kfreq - kmin)
				else
					knewfreq = kfreq
					if kcut != 1 then
						kamp = 0
					endif
				endif
			else
				if kfreq <= kmax then
					knewfreq = (kfreq - kcenter) / kfactor + kmin
				else
					knewfreq = kfreq
					if kcut != 0 then
						kamp = 0
					endif
				endif
			endif
			; Calculate the number of the new band/bin
			if knewfreq != kfreq then
				knewndx = (floor(knewfreq / ifftres) - 1)
			else
				knewndx = kndx
			endif
			knewndx limit knewndx, 0, (ifftbins -1) ; make sure we accidentally overstep the band/bin numbers
			; Write the output frequency and amplitude to the output tables
			tabw knewfreq, knewndx, ifng
			tabw kamp, knewndx, ifnb
		loop_le kndx, 1, ifftbins, loop
	endif
	; Convert the tables back to f-signals and resynthesis them
	pvsftr fsrc, ifnb, ifng
	asyn pvsynth fsrc
	aout balance asyn, adel

Output:
	xout aout
endop
